@@ -507,7 +507,7 @@ class ChoiceModel(ABC):
                 # logic for isvars
                 # Remove lambda coeffs from member class param names
                 # TODO! BELOW LINE IN PROGRESS
-                member_class_names_idx = self._get_member_X_idx(ii) #, coeff_names=self.Xnames)
+                member_class_names_idx = self._get_member_X_idx(ii, coeff_names=self.Xnames)
                 lambda_idx = np.where(np.char.find(self.Xnames, 'lambda') != -1)[0]
                 sd_idx = np.where(np.char.find(self.Xnames, 'sd') != -1)[0]
                 chol_idx = np.where(np.char.find(self.Xnames, 'chol') != -1)[0]
@@ -517,7 +517,7 @@ class ChoiceModel(ABC):
                 member_class_names_idx = np.sort(member_class_names_idx)
                 member_class_names_idx = np.array(member_class_names_idx, dtype='int32')
                 # member_class_names = self.Xnames[member_class_names_idx]
-                member_class_names = self.varnames[member_class_names_idx]
+                member_class_names = self.Xnames[member_class_names_idx]
                 if self.membership_as_probability:
                     member_class_names = ["probability"]
                 # member_class_names_idx = [np.where(np.char.find(self.Xnames, param) != -1)[0] for param in member_class]
