@@ -1,6 +1,5 @@
 """Implements Latent Class Model."""
 
-from re import A
 from .multinomial_logit import MultinomialLogit
 import numpy as np
 from scipy.optimize import minimize
@@ -27,13 +26,14 @@ class LatentClassModel(MultinomialLogit):
             ids=None, weights=None, avail=None,
             avail_latent=None,  # TODO?: separate param needed?
             transvars=None,
-            transformation=None, base_alt=None, fit_intercept=False,
+            transformation=None, base_alt=None,
             intercept_opts=None,
             init_coeff=None, init_class_betas=None, init_class_thetas=None,
-            maxiter=2000, random_state=None, ftol=1e-5,
+            maxiter=2000, random_state=None, ftol=1e-5, ftol_lccm=1e-4,
             gtol=1e-5, gtol_membership_func=1e-5, grad=True, hess=True,
             panels=None, verbose=1,
-            method="bfgs", scipy_optimisation=False):
+            method="bfgs", scipy_optimisation=False,
+            validation=False):
         """Fit multinomial and/or conditional logit models.
 
         Parameters
@@ -184,8 +184,13 @@ class LatentClassModel(MultinomialLogit):
         self.pred_prob = None
         self.pred_prob_all = None
 
-        if '_inter' in np.concatenate(class_params_spec) or '_inter' in np.concatenate(member_params_spec):
+        self.ind_pred_prob_classes = []
+        self.choice_pred_prob_classes = []
+
+        if '_inter' in np.concatenate(class_params_spec):
             fit_intercept = True
+        else:
+            fit_intercept = False
 
         # TODO?: assumes alts are nums 1...J
         if intercept_opts is None:
@@ -205,6 +210,8 @@ class LatentClassModel(MultinomialLogit):
         self.init_class_betas = init_class_betas
         self.init_class_thetas = init_class_thetas
 
+        self.validation = validation
+
         super(LatentClassModel, self).fit(X, y, varnames, alts, isvars,
                                           transvars, transformation, ids,
                                           weights, avail, base_alt,
@@ -274,6 +281,8 @@ class LatentClassModel(MultinomialLogit):
 
     def _post_fit(self, optimization_res, coeff_names, sample_size,
                   hess_inv=None, verbose=1):
+        if self.validation:
+            return
         super(LatentClassModel, self)._post_fit(optimization_res,
                                                 coeff_names,
                                                 sample_size)
@@ -431,6 +440,7 @@ class LatentClassModel(MultinomialLogit):
             ll [np.float64]: Loglik
         """
         H = self._posterior_est_latent_class_probability(class_thetas)
+        self.H = H  # save individual-level probabilities for class membership
 
         H[np.where(H < 1e-30)] = 1e-30
         weight_post = np.multiply(np.log(H), weights)
@@ -681,6 +691,8 @@ class LatentClassModel(MultinomialLogit):
 
         while not converged and iter_num < max_iter:
             # Expectation step
+            self.ind_pred_prob_classes = []
+            self.choice_pred_prob_classes = []
             X_class0_idx = self._get_class_X_idx(0)
             # Set fxidx for base class when calling multinomial loglik
             self.fxidx = class_fxidxs[0]
@@ -740,7 +752,14 @@ class LatentClassModel(MultinomialLogit):
             log_lik = np.log(np.sum(weights, axis=0))  # sum over classes
 
             log_lik_new = np.sum(log_lik)
-            if validation:
+            if self.validation:
+                self.loglikelihood = log_lik_new
+                num_params = 0
+                num_params += sum([len(betas) for betas in class_betas])
+                num_params += len(class_thetas)
+                self.aic = -2 * log_lik_new + 2 * num_params
+                self.bic = -2 * log_lik_new + num_params * np.log(self.sample_size)
+
                 return log_lik_new
 
             weights_individual = weights
@@ -792,6 +811,8 @@ class LatentClassModel(MultinomialLogit):
                     # TODO: adhoc
                     if prev_tmp_thetas_sd[ii] < 0.25 * tmp_theta_sd and prev_tmp_thetas_sd[ii] != 0.01:
                         tmp_thetas_sd[ii] = prev_tmp_thetas_sd[ii]
+                    if np.isclose(tmp_thetas_sd[ii], 1.0):
+                        tmp_thetas_sd[ii] = prev_tmp_thetas_sd[ii]
                 class_thetas_sd = tmp_thetas_sd
             else:
                 optimsation_convergences = False
@@ -828,6 +849,8 @@ class LatentClassModel(MultinomialLogit):
                 self.transvars = global_transvars
                 self.pred_prob_all = np.append(self.pred_prob_all,
                                                self.pred_prob)
+                self.ind_pred_prob_classes.append(self.ind_pred_prob)
+                self.choice_pred_prob_classes.append(self.choice_pred_prob)
                 if opt_res['success']:
                     class_betas[s] = opt_res['x']
                     prev_class_betas_sd = class_betas_sd
@@ -917,5 +940,5 @@ class LatentClassModel(MultinomialLogit):
 
     def _bfgs_optimization(self, betas, X, y, weights, avail, maxiter):
         """Override bfgs function in multinomial logit to use EM."""
-        opt_res = self._expectation_maximisation_algorithm(X, y, avail)
+        opt_res = self._expectation_maximisation_algorithm(X, y, avail, validation=self.validation)
         return opt_res
