@@ -6,7 +6,6 @@ from .boxcox_functions import boxcox_param_deriv_mixed, \
                               boxcox_transformation_mixed
 import scipy.stats as ss
 from scipy.optimize import minimize
-# from scipy.optimize import basinhopping
 from ._choice_model import ChoiceModel
 from ._device import device as dev
 import numpy as np
@@ -87,7 +86,7 @@ class MixedLogit(ChoiceModel):
             randvars=None, panels=None, base_alt=None, fit_intercept=False,
             init_coeff=None, maxiter=2000, random_state=None, correlation=None,
             n_draws=1000, halton=True, minimise_func=None, verbose=1,
-            batch_size=None, halton_opts=None, ftol=1e-8,
+            batch_size=None, halton_opts=None, ftol=1e-7,
             gtol=1e-5, hess=True, grad=True, method="bfgs",
             save_fitted_params=True):
         """Fit Mixed Logit models.
@@ -309,6 +308,7 @@ class MixedLogit(ChoiceModel):
         X = X.reshape(N, P, J, K)
         y = y.reshape(N, P, J, 1)
         self.y = y
+        self.panel_info = panel_info
 
         self.obs_prob = np.mean(np.mean(np.mean(y, axis=3), axis=1), axis=0)
 
@@ -453,6 +453,9 @@ class MixedLogit(ChoiceModel):
             H = np.linalg.inv(H)
             optimizat_res['hess_inv'] = H
 
+        if self.save_fitted_params:
+            self._save_fitted_params(self.y, self.p, panel_info, self.Br)  # TODO: TO BE REMOVED
+
         self._post_fit(optimizat_res, Xnames, N, verbose)
 
     def _reshape_weights(self, weights, panels):
@@ -517,17 +520,21 @@ class MixedLogit(ChoiceModel):
         Br = temp_br
 
         pch2 = np.multiply(pch2[:, None], Br)
-        if not dev.using_gpu:
-            pch2_res = np.zeros((self.N, self.Kr))
-            pch2_sd_test = np.zeros((self.N, self.Kr))
-            for i in range(self.N):
-                # print('pch2[i*self.n_draws:(i+1)*self.n_draws, :]', pch2[i*self.n_draws:(i+1)*self.n_draws, :])
-                pch2_res[i, :] = np.sum(pch2[i*self.batch_size:(i+1)*self.batch_size, :], axis=0)
-                pch2_sd_test[i, :] = np.std(pch2[i*self.batch_size:(i+1)*self.batch_size, :],
-                                            axis=0)
+        # if not dev.using_gpu:
+        pch2_res = np.zeros((self.N, self.Kr))
+        pch2_sd_test = np.zeros((self.N, self.Kr))
+        if dev.using_gpu:
+            pch = dev.to_gpu(pch)
+            pch2_res = dev.to_gpu(pch2_res)
+            pch2_sd_test = dev.to_gpu(pch2_sd_test)
+        for i in range(self.N):
+            # print('pch2[i*self.n_draws:(i+1)*self.n_draws, :]', pch2[i*self.n_draws:(i+1)*self.n_draws, :])
+            pch2_res[i, :] = np.sum(pch2[i*self.batch_size:(i+1)*self.batch_size, :], axis=0)
+            pch2_sd_test[i, :] = np.std(pch2[i*self.batch_size:(i+1)*self.batch_size, :],
+                                        axis=0)
 
-            self.pch2_res = pch2_res
-            self.pch2_sd_test = pch2_sd_test
+        self.pch2_res = pch2_res
+        self.pch2_sd_test = pch2_sd_test
 
     def _compute_probabilities(self, betas, X, panel_info, draws, drawstrans,
                                avail, var_list, chol_mat):
@@ -561,9 +568,10 @@ class MixedLogit(ChoiceModel):
         if dev.using_gpu:
             XBf = dev.to_gpu(XBf)
             XBr = dev.to_gpu(XBr)
+            V = dev.to_gpu(V)
         if self.Kf != 0:
             XBf = dev.cust_einsum('npjk,k -> npj', Xf, Bf)
-            # V += XBf[:, :, :, None]*self.S[:, :, :, None]
+
         if self.Kr != 0:
             # TODO? consequences of splitting R.V. and trans R.V.?
             Br = Br_b[None, :, None] + np.matmul(chol_mat[:self.Kr, :self.Kr], draws)
@@ -596,8 +604,6 @@ class MixedLogit(ChoiceModel):
             #             (1 - np.sqrt(2*(1 - draws_k)))*(draws_k > .5)
             self.Br = Br  # save Br to use later
             XBr = dev.cust_einsum('npjk,nkr -> npjr', Xr, Br)  # (N, P, J, R)
-            # TODO: Removed self.S ... any effects?
-            # print('V', type(V))
             V = XBf[:, :, :, None] + XBr
         #  transformation
         #  transformations for variables with fixed coeffs
@@ -627,26 +633,42 @@ class MixedLogit(ChoiceModel):
             V += Xbr_trans  # (N, P, J, R)
 
         #  Combine utilities of fixed and random variables
-        V[V > max_exp_val] = max_exp_val
+        # V[V > max_exp_val] = max_exp_val
         # Exponent of the utility function for the logit formula
-        eV = dev.np.exp(V)
         if avail is not None:
             if self.panels is not None:
-                eV = eV*avail[:, :, :, None]  # Acommodate availablity of alts with panels
+                V = V*avail[:, :, :, None]  # Acommodate availablity of alts with panels
             else:
-                eV = eV*avail[:, None, :, None]  # Acommodate availablity of alts.
+                V = V*avail[:, None, :, None]  # Acommodate availablity of alts.
 
+        # max_exp_val = 1400
+
+        # V[np.where(V > max_exp_val)] = max_exp_val
+        # V[np.where(V < -max_exp_val)] = -max_exp_val
+
+        # # if dev.using_gpu:
+        # #     V = dev.to_gpu(V)
+        # if dev.using_gpu:
+        #     args = (V[:, :, j, :] for j in range(self.J))
+        #     V_logsumexp = dev.np.logaddexp(*args)
+        #     V_logsumexp = dev.np.expand_dims(V_logsumexp, axis=2)
+        # else:
+        #     # TODO: idea... only do log-sum trick if overflow
+        #     V_logsumexp = dev.np.logaddexp.reduce(V, axis=2, keepdims=True)
+
+        # p = np.exp(V - V_logsumexp)
         # Thresholds to avoid overflow warnings
-        # eV[np.where(eV > max_exp_val)] = max_exp_val
-        # eV[np.where(eV < -max_exp_val)] = -max_exp_val
+        V[np.where(V > max_exp_val)] = max_exp_val
+        V[np.where(V < -max_exp_val)] = -max_exp_val
+        eV = dev.np.exp(V)
         sum_eV = dev.np.sum(eV, axis=2, keepdims=True)
         sum_eV[np.where(sum_eV < min_comp_val)] = min_comp_val
         p = np.divide(eV, sum_eV, out=np.zeros_like(eV))
-        # temp_p = np.mean(np.mean(np.mean(p, axis=0), axis=0), axis=1)
-        if panel_info is not None:
-            p = p*panel_info[:, :, None, None]
         if self.save_fitted_params:
-            self._save_fitted_params(self.y, p, panel_info, Br)  # TODO: TO BE REMOVED
+            self.p = p  # save
+        # # temp_p = np.mean(np.mean(np.mean(p, axis=0), axis=0), axis=1)
+        # if panel_info is not None:
+        #     p = p*panel_info[:, :, None, None]
         return p
 
     # # def _compute_probabilities(self, betas, X, panel_info, draws, drawstrans,
@@ -690,8 +712,6 @@ class MixedLogit(ChoiceModel):
 
         if dev.using_gpu:
             betas = dev.to_gpu(betas)
-            if hasattr(self, 'S'):
-                self.S = dev.to_gpu(self.S)
 
         beta_segment_names = ["Bf", "Br_b", "chol", "Br_w", "Bftrans",
                               "flmbda", "Brtrans_b", "Brtrans_w", "rlmda"]
@@ -784,14 +804,13 @@ class MixedLogit(ChoiceModel):
                         else:
                             corr_indices.append(Kr_all - num_rvtrans_total + rv_trans_count_all - 1)   # TODO i think
 
-
         if hasattr(self, 'correlation') and isinstance(self.correlation, bool) and self.correlation:
             corr_pairs = list(itertools.combinations(np.arange(self.Kr), 2)) + [(i, i) for i in range(self.Kr)]
         else:
-            corr_pairs = list(itertools.combinations(corr_indices, 2))
+            corr_pairs = list(itertools.combinations(corr_indices, 2)) + [(idx, idx) for ii, idx in enumerate(corr_indices)]
 
         reversed_corr_pairs = [tuple(reversed(pair)) for ii, pair in enumerate(corr_pairs)]
-        reversed_corr_pairs.sort(key= lambda x: x[0])
+        reversed_corr_pairs.sort(key=lambda x: x[0])
 
         chol_count = 0
         for _, corr_pair in enumerate(reversed_corr_pairs):
@@ -866,6 +885,10 @@ class MixedLogit(ChoiceModel):
             if dev.using_gpu:
                 draws_batch = dev.to_gpu(draws_batch)
 
+            # TODO: Put in for LCCMM get_MAE
+            self.chol_mat = chol_mat
+            self.var_list = var_list
+
             p = self._compute_probabilities(betas, X, panel_info, draws_batch, drawstrans_batch,
                                             avail, var_list, chol_mat)
             # if dev.using_gpu:
@@ -876,6 +899,7 @@ class MixedLogit(ChoiceModel):
             self.choice_pred_prob = np.mean(p, axis=3)
             self.ind_pred_prob = np.mean(self.choice_pred_prob, axis=1)
             self.pred_prob = np.mean(self.ind_pred_prob, axis=0)
+            self.prob_full = p
             # self.obs_prob = np.mean(np.mean(np.mean(y, axis=3), axis=1), axis=0)
 
             # Joint probability estimation for panels data
@@ -1049,6 +1073,8 @@ class MixedLogit(ChoiceModel):
         g = g/lik[:, None]
         # updated gradient
         if weights is not None:
+            if weights.ndim == 1:
+                g = np.transpose(np.transpose(g) * weights)
             if weights.ndim == 2:
                 g = np.transpose(np.transpose(g)*weights[:, 0])
             if weights.ndim == 3:  # (3 dim for panel data)
@@ -1064,8 +1090,8 @@ class MixedLogit(ChoiceModel):
         if self.grad:
             result = (-loglik, -g)
 
-        if self.verbose > 1:
-            print(f"Evaluation {self.total_fun_eval} Log-Lik.={loglik:.2f}")
+        # if self.verbose > 1:
+        #     print(f"Evaluation {self.total_fun_eval} Log-Lik.={loglik:.3f}")
 
         return result
 
@@ -1110,11 +1136,6 @@ class MixedLogit(ChoiceModel):
 
         self.y = validation_Y
 
-        S = np.zeros((N, self.P, self.J))
-        for i in range(N):
-            S[i, 0:self.P_i[i], :] = 1
-        self.S = S
-
         res = self._loglik_gradient(betas, validation_X, validation_Y,
                                     avail=avail, weights=weights,
                                     panel_info=panel_info,
@@ -1122,7 +1143,7 @@ class MixedLogit(ChoiceModel):
                                     batch_size=self.batch_size
                                     )
         loglik = -res[0]
-        print('Validation loglik: ', loglik)
+        # print('Validation loglik: ', loglik)
         return loglik
 
     def _prob_product_across_panels(self, pch, panel_info):
